"""Module with attributes defining month ranges for the four seasons.

Used for convenience with Climatology.multiyear_averages().
Note: Summer is defined as JJAS, as opposed to the meteorological seasons of JJA.
"""

SPRING = [(3, 5)]
SUMMER = [(6, 9)]
AUTUMN = [(10, 11)]
WINTER = [(12, 2)]
ALL = [(3, 5), (6, 9), (10, 11), (12, 2)]
